# ruff: noqa: PLR2004, ERA001

import logging
from typing import Any, Type

import pytest
from gwproto import MQTTTopic
from paho.mqtt.client import MQTT_ERR_CONN_LOST

from gwproactor.links import StateName
from gwproactor_test.certs import uses_tls
from gwproactor_test.comm_test_helper import CommTestHelper
from gwproactor_test.wait import await_for


@pytest.mark.asyncio
class ProactorCommBasicTests:
    CTH: Type[CommTestHelper]

    async def test_no_parent(self) -> None:
        async with self.CTH(add_child=True) as h:
            child = h.child
            stats = child.stats.link(child.upstream_client)
            comm_event_counts = stats.comm_event_counts
            link = child.links.link(child.upstream_client)

            # unstarted child
            assert stats.num_received == 0
            assert link.state == StateName.not_started
            child.logger.info(child.settings.model_dump_json(indent=2))

            # start child
            h.start_child()
            await await_for(
                link.active_for_send,
                1,
                "ERROR waiting link active_for_send",
                err_str_f=child.summary_str,
            )
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_peer
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 1
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert len(stats.comm_events) == 2
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child.event_persister

            # Tell client we lost comm.
            child.mqtt_client_wrapper("gridworks").mqtt_client._loop_rc_handle(  # noqa: SLF001
                MQTT_ERR_CONN_LOST
            )

            # Wait for reconnect
            await await_for(
                lambda: stats.comm_event_counts[
                    "gridworks.event.comm.mqtt.fully.subscribed"
                ]
                > 1,
                3,
                "ERROR waiting link to resubscribe after comm loss",
                err_str_f=child.summary_str,
            )
            assert link.active_for_send()
            assert not link.active_for_recv()
            assert not link.active()
            assert link.state == StateName.awaiting_peer
            assert comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"] == 2
            assert comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert len(stats.comm_events) == 5
            for comm_event in stats.comm_events:
                assert comm_event.MessageId in child.event_persister

    async def test_basic_comm_child_first(self) -> None:
        async with self.CTH(add_child=True, add_parent=True) as h:
            child = h.child
            child_stats = child.stats.link(child.upstream_client)
            child_comm_event_counts = child_stats.comm_event_counts
            child_link = child.links.link(child.upstream_client)

            # unstarted child, parent
            assert child_stats.num_received == 0
            assert child_link.state == StateName.not_started

            # start child
            h.start_child()
            await await_for(
                child_link.active_for_send,
                1,
                "ERROR waiting link active_for_send",
                err_str_f=child.summary_str,
            )
            assert not child_link.active_for_recv()
            assert not child_link.active()
            assert child_link.state == StateName.awaiting_peer
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 1
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 0
            assert len(child_stats.comm_events) == 2
            for comm_event in child_stats.comm_events:
                assert comm_event.MessageId in child.event_persister

            # start parent
            h.start_parent()

            # wait for link to go active
            await await_for(
                child_link.active,
                10,
                "ERROR waiting link active",
                err_str_f=child.summary_str,
            )
            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 1
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 1
            assert len(child_stats.comm_events) == 3

            # wait for all events to be acked
            await await_for(
                lambda: child.event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

            # Tell client we lost comm.
            child.mqtt_client_wrapper("gridworks").mqtt_client._loop_rc_handle(  # noqa: SLF001
                MQTT_ERR_CONN_LOST
            )

            # Wait for reconnect
            await await_for(
                lambda: child_stats.comm_event_counts[
                    "gridworks.event.comm.peer.active"
                ]
                > 1,
                3,
                "ERROR waiting link to resubscribe after comm loss",
                err_str_f=child.summary_str,
            )
            assert child_link.active_for_send()
            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 2
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 2
            assert len(child_stats.comm_events) == 7

            # wait for all events to be acked
            await await_for(
                lambda: child.event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

    # @pytest.mark.asyncio
    # async def test_broker_dns_failure(self):
    #     """Verify proactor does not crash in presence of bad host.
    #
    #     This test commented out because it's very slow - 4 seconds before mqtt client thread
    #     reports first problem.
    #
    #     A better test would:
    #         - be faster
    #         - verify client thread continues
    #         - verify problem reports occur
    #         - possibly verify that eventually child requests device restart since pi's are
    #           having DNS failures on network comm loss that seem to be undone by pi restart
    #     """
    #     no_broker = MQTTClient(host="www.foo.com")
    #     async with self.CTH(
    #         child_settings=DummyChildSettings(
    #             parent_mqtt=no_broker,
    #         ),
    #         start_child=True,
    #         verbose=True,
    #     ):
    #         for i in range(20):
    #             print(f"{i}...")
    #             await asyncio.sleep(1)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("suppress_tls", [False, True])
    async def test_basic_comm_parent_first(
        self, request: Any, suppress_tls: bool
    ) -> None:
        async with self.CTH() as h:
            base_logger = logging.getLogger(
                h.child_helper.settings.logging.base_log_name
            )
            base_logger.warning(f"{request.node.name}  suppress_tls: {suppress_tls}")
            if suppress_tls:
                if not uses_tls(h.child_helper.settings) and not uses_tls(
                    h.parent_helper.settings
                ):
                    base_logger.warning(
                        "Skipping test <%s> since TLS has already been suppressed by environment variables",
                        request.node.name,
                    )
                else:
                    h.set_use_tls(False)
            h.add_child()
            h.add_parent()
            child = h.child
            child_stats = child.stats.link(child.upstream_client)
            child_comm_event_counts = child_stats.comm_event_counts
            child_link = child.links.link(child.upstream_client)
            parent = h.parent
            parent_link = parent.links.link(parent.downstream_client)

            # unstarted parent
            assert parent_link.state == StateName.not_started

            # start parent
            h.start_parent()
            await await_for(
                parent_link.active_for_send,
                1,
                "ERROR waiting link active_for_send",
                err_str_f=parent.summary_str,
            )

            # unstarted child
            assert child_stats.num_received == 0
            assert child_link.state == StateName.not_started

            # start child
            h.start_child()
            await await_for(
                child_link.active,
                1,
                "ERROR waiting link active",
                err_str_f=parent.summary_str,
            )

            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 1
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 1
            assert len(child_stats.comm_events) == 3

            # wait for all events to be acked
            await await_for(
                lambda: child.event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

    @pytest.mark.asyncio
    async def test_basic_parent_comm_loss(self) -> None:
        async with self.CTH(add_child=True, add_parent=True, verbose=False) as h:
            child = h.child
            child_stats = child.stats.link(child.upstream_client)
            child_comm_event_counts = child_stats.comm_event_counts
            child_link = child.links.link(child.upstream_client)
            parent = h.parent
            parent_link = parent.links.link(parent.downstream_client)

            # unstarted child, parent
            assert parent_link.state == StateName.not_started
            assert child_stats.num_received == 0
            assert child_link.state == StateName.not_started

            # start child, parent
            h.start_child()
            h.start_parent()
            await await_for(
                child_link.active,
                1,
                "ERROR waiting link active",
                err_str_f=child.summary_str,
            )
            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 1
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 1
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 0
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 1
            assert len(child_stats.comm_events) == 3

            # wait for all events to be acked
            await await_for(
                lambda: child.event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

            # Tell *child* client we lost comm.
            child.mqtt_client_wrapper(  # noqa: SLF001
                child.upstream_client
            ).mqtt_client._loop_rc_handle(MQTT_ERR_CONN_LOST)

            # Wait for reconnect
            await await_for(
                lambda: child_stats.comm_event_counts[
                    "gridworks.event.comm.peer.active"
                ]
                > 1,
                3,
                "ERROR waiting link to resubscribe after comm loss",
                err_str_f=child.summary_str,
            )
            assert child_link.active_for_send()
            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 2
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 2
            assert len(child_stats.comm_events) == 7

            # wait for all events to be acked
            await await_for(
                lambda: child.event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )

            # Tell *parent* client we lost comm.
            parent.mqtt_client_wrapper(  # noqa: SLF001
                parent.downstream_client
            ).mqtt_client._loop_rc_handle(MQTT_ERR_CONN_LOST)
            # wait for child to get ping from parent when parent reconnects to mqtt
            # noinspection PyTypeChecker
            parent_ping_topic = MQTTTopic.encode(
                envelope_type="gw",
                src=parent.publication_name,
                dst=child.subscription_name,
                message_type="gridworks-ping",
            )
            num_parent_pings = child_stats.num_received_by_topic[parent_ping_topic]
            await await_for(
                lambda: child_stats.num_received_by_topic[parent_ping_topic]
                == num_parent_pings + 1,
                3,
                f"ERROR waiting for parent ping {parent_ping_topic}",
                err_str_f=child.summary_str,
            )
            # verify no child comm state change has occurred.
            err_str = f"\n{child.summary_str()}\n" f"{parent.summary_str()}\n"
            assert child_link.active_for_send()
            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active, err_str
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 2
            ), err_str
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 2
            ), err_str
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 1
            ), err_str
            assert (
                child_comm_event_counts["gridworks.event.comm.peer.active"] == 2
            ), err_str
            assert len(child_stats.comm_events) == 7, err_str
            assert child.event_persister.num_pending == 0, err_str

            # Tell *both* clients we lost comm.
            parent.mqtt_client_wrapper(  # noqa: SLF001
                parent.downstream_client
            ).mqtt_client._loop_rc_handle(MQTT_ERR_CONN_LOST)
            child.mqtt_client_wrapper(  # noqa: SLF001
                child.upstream_client
            ).mqtt_client._loop_rc_handle(MQTT_ERR_CONN_LOST)

            # Wait for reconnect
            await await_for(
                lambda: child_stats.comm_event_counts[
                    "gridworks.event.comm.peer.active"
                ]
                > 2,
                3,
                "ERROR waiting link to resubscribe after comm loss",
                err_str_f=child.summary_str,
            )
            assert child_link.active_for_send()
            assert child_link.active_for_recv()
            assert child_link.active()
            assert child_link.state == StateName.active
            assert child_comm_event_counts["gridworks.event.comm.mqtt.connect"] == 3
            assert (
                child_comm_event_counts["gridworks.event.comm.mqtt.fully.subscribed"]
                == 3
            )
            assert child_comm_event_counts["gridworks.event.comm.mqtt.disconnect"] == 2
            assert child_comm_event_counts["gridworks.event.comm.peer.active"] == 3
            assert len(child_stats.comm_events) == 11

            # wait for all events to be acked
            await await_for(
                lambda: child.event_persister.num_pending == 0,
                1,
                "ERROR waiting for events to be acked",
                err_str_f=child.summary_str,
            )



"""Scada implementation"""

from typing import Any, Optional, cast

from gwproto import Message
from gwproto.messages import EventBase

from gwproactor.links.link_settings import LinkSettings
from gwproactor.message import MQTTReceiptPayload
from gwproactor.persister import SimpleDirectoryWriter
from gwproactor.proactor_implementation import Proactor
from gwproactor_test.dummies import DUMMY_ATN_NAME
from gwproactor_test.dummies.names import DUMMY_ATN_SHORT_NAME
from gwproactor_test.dummies.tree.atn_settings import DummyAtnSettings
from gwproactor_test.dummies.tree.codecs import DummyCodec


class DummyAtn(Proactor):
    def __init__(
        self,
        name: str = "",
        settings: Optional[DummyAtnSettings] = None,
    ) -> None:
        super().__init__(
            name=name or DUMMY_ATN_NAME,
            settings=DummyAtnSettings() if settings is None else settings,
        )
        self._links.add_mqtt_link(
            LinkSettings(
                client_name=self.settings.scada1_link.client_name,
                gnode_name=self.settings.scada1_link.long_name,
                spaceheat_name=self.settings.scada1_link.short_name,
                mqtt=self.settings.scada1_link,
                codec=DummyCodec(
                    src_name=self.settings.scada1_link.long_name,
                    dst_name=DUMMY_ATN_SHORT_NAME,
                    model_name="Scada1ToAtn1Message",
                ),
                downstream=True,
            ),
        )
        self.links.log_subscriptions()

    @classmethod
    def make_event_persister(cls, settings: DummyAtnSettings) -> SimpleDirectoryWriter:
        return SimpleDirectoryWriter(settings.paths.event_dir)

    @property
    def publication_name(self) -> str:
        return self.name

    @property
    def subscription_name(self) -> str:
        return "a"

    @property
    def settings(self) -> DummyAtnSettings:
        return cast(DummyAtnSettings, self._settings)

    def _derived_process_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[Any]
    ) -> None:
        self._logger.path(
            f"++{self.name}._derived_process_mqtt_message %s",
            message.Payload.message.topic,
        )
        path_dbg = 0
        self.stats.add_message(decoded)
        match decoded.Payload:
            case EventBase():
                path_dbg |= 0x00000008
            case _:
                path_dbg |= 0x00000040
        self._logger.path(
            f"--{self.name}._derived_process_mqtt_message  path:0x%08X", path_dbg
        )



import abc
import enum
from dataclasses import dataclass
from typing import Optional, Sequence

from gwproto import Message
from result import Err, Ok, Result

from gwproactor.message import (
    MQTTConnectFailPayload,
    MQTTConnectPayload,
    MQTTDisconnectPayload,
    MQTTReceiptPayload,
)


class StateName(enum.Enum):
    none = "none"
    not_started = "not_started"
    connecting = "connecting"
    awaiting_setup_and_peer = "awaiting_setup_and_peer"
    awaiting_setup = "awaiting_setup"
    awaiting_peer = "awaiting_peer"
    active = "active"
    stopped = "stopped"


def state_is_active(state: StateName) -> bool:
    return state == StateName.active


def state_is_active_for_send(state: StateName) -> bool:
    return state in {StateName.active, StateName.awaiting_peer}


def state_is_active_for_recv(state: StateName) -> bool:
    return state_is_active(state)


class TransitionName(enum.Enum):
    none = "none"
    start_called = "start_called"
    mqtt_connected = "mqtt_connected"
    mqtt_connect_failed = "mqtt_connect_failed"
    mqtt_disconnected = "mqtt_disconnected"
    mqtt_suback = "mqtt_suback"
    message_from_peer = "message_from_peer"
    response_timeout = "response_timeout"
    stop_called = "stop_called"


@dataclass
class Transition:
    link_name: str = ""
    transition_name: TransitionName = TransitionName.none
    old_state: StateName = StateName.not_started
    new_state: StateName = StateName.not_started

    def __str__(self) -> str:
        return f"{self.link_name}:  {self.old_state.value} -- {self.transition_name.value} --> {self.new_state.value}"

    def __bool__(self) -> bool:
        return self.old_state != self.new_state

    def active(self) -> bool:
        return state_is_active(self.new_state)

    def activated(self) -> bool:
        return not state_is_active(self.old_state) and state_is_active(self.new_state)

    def deactivated(self) -> bool:
        return state_is_active(self.old_state) and not state_is_active(self.new_state)

    def send_is_active(self) -> bool:
        return state_is_active_for_send(self.new_state)

    def send_activated(self) -> bool:
        return self.send_is_active() and not state_is_active_for_send(self.old_state)

    def send_deactivated(self) -> bool:
        return not self.send_is_active() and state_is_active_for_send(self.old_state)

    def recv_activated(self) -> bool:
        return self.activated()

    def recv_deactivated(self) -> bool:
        return self.deactivated()


class InvalidCommStateInput(Exception):
    name: str = ""
    current_state: StateName = StateName.none
    transition: TransitionName = TransitionName.none

    def __init__(
        self,
        name: str = "",
        current_state: StateName = StateName.none,
        transition: TransitionName = TransitionName.none,
        *,
        msg: str = "",
    ) -> None:
        self.name = name
        self.current_state = current_state
        self.transition = transition
        super().__init__(msg)

    def __str__(self) -> str:
        s = self.__class__.__name__
        super_str = super().__str__()
        if super_str:
            s += f" [{super_str}]"
        s += f"  for link: [{self.name}]  current state:{self.current_state}  requested transition: {self.transition}"
        return s


class CommLinkMissing(InvalidCommStateInput):
    def __init__(self, name: str, *, msg: str = "") -> None:
        super().__init__(name, msg=msg)


class CommLinkAlreadyExists(InvalidCommStateInput): ...


class RuntimeLinkStateError(InvalidCommStateInput): ...


class State(abc.ABC):
    """By default all transitions disallowed except stopping, which is always allowed and leads to stopped."""

    @property
    @abc.abstractmethod
    def name(self) -> StateName:
        raise NotImplementedError

    def start(self) -> Result[Transition, InvalidCommStateInput]:
        return Err(
            InvalidCommStateInput(
                "", current_state=self.name, transition=TransitionName.start_called
            )
        )

    def stop(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition("", TransitionName.stop_called, self.name, StateName.stopped)
        )

    def process_mqtt_connected(self) -> Result[Transition, InvalidCommStateInput]:
        return Err(
            InvalidCommStateInput(
                "", current_state=self.name, transition=TransitionName.mqtt_connected
            )
        )

    def process_mqtt_disconnected(self) -> Result[Transition, InvalidCommStateInput]:
        return Err(
            InvalidCommStateInput(
                "", current_state=self.name, transition=TransitionName.mqtt_disconnected
            )
        )

    def process_mqtt_connect_fail(self) -> Result[Transition, InvalidCommStateInput]:
        return Err(
            InvalidCommStateInput(
                "",
                current_state=self.name,
                transition=TransitionName.mqtt_connect_failed,
            )
        )

    def process_mqtt_suback(
        self,
        num_pending_subscriptions: int,  # noqa: ARG002
    ) -> Result[Transition, InvalidCommStateInput]:
        return Err(
            InvalidCommStateInput(
                "", current_state=self.name, transition=TransitionName.mqtt_suback
            )
        )

    def process_mqtt_message(self) -> Result[Transition, InvalidCommStateInput]:
        return Err(
            InvalidCommStateInput(
                "", current_state=self.name, transition=TransitionName.message_from_peer
            )
        )

    def process_ack_timeout(self) -> Result[Transition, InvalidCommStateInput]:
        return Err(
            InvalidCommStateInput(
                "", current_state=self.name, transition=TransitionName.response_timeout
            )
        )

    def active(self) -> bool:
        return state_is_active(self.name)

    def active_for_send(self) -> bool:
        return state_is_active_for_send(self.name)

    def active_for_recv(self) -> bool:
        return state_is_active_for_recv(self.name)


class NotStarted(State):
    @property
    def name(self) -> StateName:
        return StateName.not_started

    def start(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition("", TransitionName.start_called, self.name, StateName.connecting)
        )


class Connecting(State):
    @property
    def name(self) -> StateName:
        return StateName.connecting

    def process_mqtt_connected(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition(
                "",
                TransitionName.mqtt_connected,
                self.name,
                StateName.awaiting_setup_and_peer,
            )
        )

    def process_mqtt_connect_fail(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition(
                "", TransitionName.mqtt_connect_failed, self.name, StateName.connecting
            )
        )


class AwaitingSetupAndPeer(State):
    @property
    def name(self) -> StateName:
        return StateName.awaiting_setup_and_peer

    def process_mqtt_disconnected(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition(
                "", TransitionName.mqtt_disconnected, self.name, StateName.connecting
            )
        )

    def process_mqtt_suback(
        self, num_pending_subscriptions: int
    ) -> Result[Transition, InvalidCommStateInput]:
        if num_pending_subscriptions == 0:
            new_state = StateName.awaiting_peer
        else:
            new_state = self.name
        return Ok(Transition("", TransitionName.mqtt_suback, self.name, new_state))

    def process_mqtt_message(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition(
                "",
                TransitionName.message_from_peer,
                self.name,
                StateName.awaiting_setup,
            )
        )


class AwaitingSetup(State):
    @property
    def name(self) -> StateName:
        return StateName.awaiting_setup

    def process_mqtt_disconnected(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition(
                "", TransitionName.mqtt_disconnected, self.name, StateName.connecting
            )
        )

    def process_mqtt_suback(
        self, num_pending_subscriptions: int
    ) -> Result[Transition, InvalidCommStateInput]:
        new_state = StateName.active if num_pending_subscriptions == 0 else self.name
        return Ok(Transition("", TransitionName.mqtt_suback, self.name, new_state))

    def process_mqtt_message(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition("", TransitionName.message_from_peer, self.name, self.name)
        )


class AwaitingPeer(State):
    @property
    def name(self) -> StateName:
        return StateName.awaiting_peer

    def process_mqtt_disconnected(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition(
                "", TransitionName.mqtt_disconnected, self.name, StateName.connecting
            )
        )

    def process_mqtt_message(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition(
                "", TransitionName.message_from_peer, self.name, StateName.active
            )
        )

    def process_ack_timeout(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(Transition("", TransitionName.response_timeout, self.name, self.name))


class Active(State):
    @property
    def name(self) -> StateName:
        return StateName.active

    def process_mqtt_disconnected(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition(
                "", TransitionName.mqtt_disconnected, self.name, StateName.connecting
            )
        )

    def process_mqtt_message(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition(
                "", TransitionName.message_from_peer, self.name, StateName.active
            )
        )

    def process_ack_timeout(self) -> Result[Transition, InvalidCommStateInput]:
        return Ok(
            Transition(
                "", TransitionName.response_timeout, self.name, StateName.awaiting_peer
            )
        )


class Stopped(State):
    @property
    def name(self) -> StateName:
        return StateName.stopped


class LinkState:
    name: str
    states: dict[StateName, State]
    curr_state: State

    def __init__(
        self, name: str, curr_state: StateName = StateName.not_started
    ) -> None:
        self.name = name
        self.states = {
            state.name: state
            for state in [
                NotStarted(),
                Connecting(),
                AwaitingSetupAndPeer(),
                AwaitingSetup(),
                AwaitingPeer(),
                Active(),
                Stopped(),
            ]
        }
        self.curr_state = self.states[curr_state]

    @property
    def state(self) -> StateName:
        return self.curr_state.name

    def in_state(self, state: StateName) -> bool:
        return self.state == state

    def active(self) -> bool:
        return self.curr_state.active()

    def active_for_send(self) -> bool:
        return self.curr_state.active_for_send()

    def active_for_recv(self) -> bool:
        return self.curr_state.active_for_recv()

    def _handle(self, result: Result) -> Result[Transition, InvalidCommStateInput]:
        match result:
            case Ok(transition):
                transition.link_name = self.name
                self.curr_state = self.states[transition.new_state]
            case Err(exception):
                exception.name = self.name
        return result

    def start(self) -> Result[Transition, InvalidCommStateInput]:
        return self._handle(self.curr_state.start())

    def stop(self) -> Result[Transition, InvalidCommStateInput]:
        return self._handle(self.curr_state.stop())

    def process_mqtt_connected(self) -> Result[Transition, InvalidCommStateInput]:
        return self._handle(self.curr_state.process_mqtt_connected())

    def process_mqtt_disconnected(self) -> Result[Transition, InvalidCommStateInput]:
        return self._handle(self.curr_state.process_mqtt_disconnected())

    def process_mqtt_connect_fail(self) -> Result[Transition, InvalidCommStateInput]:
        return self._handle(self.curr_state.process_mqtt_connect_fail())

    def process_mqtt_suback(
        self, num_pending_subscriptions: int
    ) -> Result[Transition, InvalidCommStateInput]:
        return self._handle(
            self.curr_state.process_mqtt_suback(num_pending_subscriptions)
        )

    def process_mqtt_message(self) -> Result[Transition, InvalidCommStateInput]:
        return self._handle(self.curr_state.process_mqtt_message())

    def process_ack_timeout(self) -> Result[Transition, InvalidCommStateInput]:
        return self._handle(self.curr_state.process_ack_timeout())


class LinkStates:
    _links: dict[str, LinkState]

    def __init__(self, names: Optional[Sequence[str]] = None) -> None:
        self._links = {}
        if names is not None:
            for name in names:
                self.add(name)

    def link(self, name: str) -> Optional[LinkState]:
        return self._links.get(name, None)

    def link_state(self, name: str) -> Optional[StateName]:
        if name in self._links:
            return self._links[name].curr_state.name
        return None

    def stopped(self, name: str) -> bool:
        return self.link_state(name) == StateName.stopped

    def __contains__(self, name: str) -> bool:
        return name in self._links

    def __getitem__(self, name: str) -> LinkState:
        try:
            return self._links[name]
        except KeyError as e:
            raise CommLinkMissing(name) from e

    def link_names(self) -> list[str]:
        return list(self._links.keys())

    def add(self, name: str, state: StateName = StateName.not_started) -> LinkState:
        if name in self._links:
            raise CommLinkAlreadyExists(
                name, current_state=self._links[name].curr_state.name
            )
        self._links[name] = LinkState(name, state)
        return self._links[name]

    def start_all(self) -> Result[bool, Sequence[Exception]]:
        errors = []
        for link in self._links.values():
            match link.start():
                case Err(exception):
                    errors.append(exception)
        if errors:
            return Err(errors)
        return Ok()

    def start(self, name: str) -> Result[Transition, InvalidCommStateInput]:
        return self[name].start()

    def stop(self, name: str) -> Result[Transition, InvalidCommStateInput]:
        return self[name].stop()

    def process_mqtt_connected(
        self, message: Message[MQTTConnectPayload]
    ) -> Result[Transition, InvalidCommStateInput]:
        return self[message.Payload.client_name].process_mqtt_connected()

    def process_mqtt_disconnected(
        self, message: Message[MQTTDisconnectPayload]
    ) -> Result[Transition, InvalidCommStateInput]:
        return self[message.Payload.client_name].process_mqtt_disconnected()

    def process_mqtt_connect_fail(
        self, message: Message[MQTTConnectFailPayload]
    ) -> Result[Transition, InvalidCommStateInput]:
        return self[message.Payload.client_name].process_mqtt_connect_fail()

    def process_mqtt_suback(
        self, name: str, num_pending_subscriptions: int
    ) -> Result[Transition, InvalidCommStateInput]:
        return self[name].process_mqtt_suback(num_pending_subscriptions)

    def process_mqtt_message(
        self, message: Message[MQTTReceiptPayload]
    ) -> Result[Transition, InvalidCommStateInput]:
        return self[message.Payload.client_name].process_mqtt_message()

    def process_ack_timeout(
        self, name: str
    ) -> Result[Transition, InvalidCommStateInput]:
        return self[name].process_ack_timeout()



import abc
import contextlib
import datetime
import re
import shutil
import subprocess
import time
from abc import abstractmethod
from pathlib import Path
from typing import NamedTuple, Optional

from result import Err, Ok, Result

from gwproactor.problems import Problems


class PersisterException(Exception):
    path: Optional[Path] = None
    uid: str = ""

    def __init__(
        self, msg: str = "", uid: str = "", path: Optional[Path] = None
    ) -> None:
        self.path = path
        self.uid = uid
        super().__init__(msg)

    def __str__(self) -> str:
        s = self.__class__.__name__
        super_str = super().__str__()
        if super_str:
            s += f" [{super_str}]"
        s += f"  for uid: {self.uid}  path:{self.path}"
        return s


class PersisterError(PersisterException): ...


class PersisterWarning(PersisterException): ...


class WriteFailed(PersisterError): ...


class ContentTooLarge(PersisterError): ...


class FileMissing(PersisterError): ...


class ReadFailed(PersisterError): ...


class TrimFailed(PersisterError): ...


class ReindexError(PersisterError): ...


class DecodingError(PersisterError): ...


class ByteDecodingError(DecodingError): ...


class JSONDecodingError(DecodingError): ...


class UIDExistedWarning(PersisterWarning): ...


class FileExistedWarning(PersisterWarning): ...


class FileMissingWarning(PersisterWarning): ...


class UIDMissingWarning(PersisterWarning): ...


class FileEmptyWarning(PersisterWarning): ...


class PersisterInterface(abc.ABC):
    @abstractmethod
    def persist(self, uid: str, content: bytes) -> Result[bool, Problems]:
        """Persist content, indexed by uid"""

    @abstractmethod
    def clear(self, uid: str) -> Result[bool, Problems]:
        """Delete content persisted for uid. It is error to clear a uid which is not currently persisted."""

    @abstractmethod
    def pending_ids(self) -> list[str]:
        """Get list of pending (persisted and not cleared) uids"""

    @property
    @abstractmethod
    def num_pending(self) -> int:
        """Get number of pending uids"""

    @property
    @abstractmethod
    def curr_bytes(self) -> int:
        """Return number of bytes used to store events, if known."""

    @abstractmethod
    def __contains__(self, uid: str) -> bool:
        """Check whether a uid is pending"""

    @abstractmethod
    def retrieve(self, uid: str) -> Result[Optional[bytes], Problems]:
        """Load and return persisted content for uid"""

    @abstractmethod
    def reindex(self) -> Result[Optional[bool], Problems]:
        """Re-created pending index from persisted storage"""


class _PersistedItem(NamedTuple):
    uid: str
    path: Path


class StubPersister(PersisterInterface):
    def persist(self, uid: str, content: bytes) -> Result[bool, Problems]:  # noqa: ARG002
        return Ok()

    def clear(self, uid: str) -> Result[bool, Problems]:  # noqa: ARG002
        return Ok()

    def pending_ids(self) -> list[str]:
        return []

    @property
    def num_pending(self) -> int:
        return 0

    @property
    def curr_bytes(self) -> int:
        return 0

    def __contains__(self, uid: str) -> bool:
        return False

    def retrieve(self, uid: str) -> Result[Optional[bytes], Problems]:  # noqa: ARG002
        return Ok(None)

    def reindex(self) -> Result[Optional[bool], Problems]:
        return Ok()


class SimpleDirectoryWriter(StubPersister):
    _base_dir: Path

    def __init__(
        self,
        base_dir: Path | str,
    ) -> None:
        self._base_dir = Path(base_dir).resolve()

    @classmethod
    def _make_name(cls, dt: datetime, uid: str) -> str:
        return f"{dt.isoformat()}.uid[{uid}].json"

    def persist(self, uid: str, content: bytes) -> Result[bool, Problems]:
        problems = Problems()
        try:
            if not self._base_dir.exists():
                self._base_dir.mkdir(parents=True, exist_ok=True)
            path = self._base_dir / self._make_name(
                datetime.datetime.now(tz=datetime.timezone.utc), uid
            )
            try:
                with path.open("wb") as f:
                    f.write(content)
            except Exception as e:  # pragma: no cover  # noqa: BLE001
                return Err(
                    problems.add_error(e).add_error(
                        WriteFailed("Open or write failed", uid=uid, path=path)
                    )
                )
        except Exception as e:  # noqa: BLE001
            return Err(
                problems.add_error(e).add_error(
                    PersisterError("Unexpected error", uid=uid)
                )
            )
        if problems:
            return Err(problems)
        return Ok()


class TimedRollingFilePersister(PersisterInterface):
    DEFAULT_MAX_BYTES: int = 500 * 1024 * 1024
    FILENAME_RGX: re.Pattern = re.compile(r"(?P<dt>.*)\.uid\[(?P<uid>.*)].json$")
    REINDEX_PAT_SECONDS = 1.0

    _base_dir: Path
    _max_bytes: int = DEFAULT_MAX_BYTES
    _pending: dict[str, Path]
    _curr_dir: Path
    _curr_bytes: int
    _pat_watchdog_args: Optional[list[str]] = None
    _reindex_pat_seconds: float = REINDEX_PAT_SECONDS

    def __init__(
        self,
        base_dir: Path | str,
        max_bytes: int = DEFAULT_MAX_BYTES,
        pat_watchdog_args: Optional[list[str]] = None,
        reindex_pat_seconds: float = REINDEX_PAT_SECONDS,
    ) -> None:
        self._base_dir = Path(base_dir).resolve()
        self._max_bytes = max_bytes
        self._curr_dir = self._today_dir()
        self._curr_bytes = 0
        self._pat_watchdog_args = pat_watchdog_args
        self._reindex_pat_seconds = reindex_pat_seconds
        self._pending = {}

    @property
    def max_bytes(self) -> int:
        return self._max_bytes

    @property
    def curr_bytes(self) -> int:
        return self._curr_bytes

    @property
    def base_dir(self) -> Path:
        return self._base_dir

    @property
    def curr_dir(self) -> Path:
        return self._curr_dir

    def persist(self, uid: str, content: bytes) -> Result[bool, Problems]:
        problems = Problems()
        try:
            if len(content) > self._max_bytes:
                return Err(
                    problems.add_error(
                        ContentTooLarge(
                            f"content bytes ({len(content)} > max bytes {self._max_bytes}",
                            uid=uid,
                        )
                    )
                )
            if len(content) + self._curr_bytes > self._max_bytes:
                trimmed = self._trim_old_storage(len(content))
                match trimmed:
                    case Err(trim_problems):
                        problems.add_problems(trim_problems)
                        if problems.errors:
                            return Err(problems.add_error(TrimFailed(uid=uid)))
            existing_path = self._pending.pop(uid, None)
            if existing_path is not None:
                problems.add_warning(UIDExistedWarning(uid=uid, path=existing_path))
                if existing_path.exists():
                    self._curr_bytes -= existing_path.stat().st_size
                    problems.add_warning(
                        FileExistedWarning(uid=uid, path=existing_path)
                    )
                else:
                    problems.add_warning(
                        FileMissingWarning(uid=uid, path=existing_path)
                    )
            self._roll_curr_dir()
            self._pending[uid] = self._curr_dir / self._make_name(
                datetime.datetime.now(tz=datetime.timezone.utc), uid
            )
            try:
                with self._pending[uid].open("wb") as f:
                    f.write(content)
                self._curr_bytes += len(content)
            except Exception as e:  # pragma: no cover  # noqa: BLE001
                return Err(
                    problems.add_error(e).add_error(
                        WriteFailed("Open or write failed", uid=uid, path=existing_path)
                    )
                )
        except Exception as e:  # noqa: BLE001
            return Err(
                problems.add_error(e).add_error(
                    PersisterError("Unexpected error", uid=uid)
                )
            )
        if problems:
            return Err(problems)
        return Ok()

    def _trim_old_storage(self, needed_bytes: int) -> Result[bool, Problems]:
        problems = Problems()
        last_day_dir: Optional[Path] = None
        items = list(self._pending.items())
        for uid, path in items:
            try:
                match self.clear(uid):
                    case Err(other):
                        problems.add_problems(other)
                day_dir = path.parent
                if last_day_dir is not None and last_day_dir != day_dir:
                    shutil.rmtree(last_day_dir, ignore_errors=True)
                last_day_dir = day_dir
            except Exception as e:  # noqa: BLE001
                problems.add_error(e)
                problems.add_error(
                    PersisterError("Unexpected error", uid=uid, path=path)
                )
            if self._curr_bytes <= self._max_bytes - needed_bytes:
                break
        try:
            if last_day_dir is not None and (
                not self._pending
                or next(iter(self._pending.values())).parent != last_day_dir
            ):
                shutil.rmtree(last_day_dir, ignore_errors=True)
        except Exception as e:  # pragma: no cover  # noqa: BLE001
            problems.add_error(e)
            problems.add_error(PersisterError("Unexpected error"))
        if problems:
            return Err(problems)
        return Ok()

    def clear(self, uid: str) -> Result[bool, Problems]:
        problems = Problems()
        path = self._pending.pop(uid, None)
        if path:
            if path.exists():
                self._curr_bytes -= path.stat().st_size
                path.unlink()
                path_dir = path.parent
                # Remove directory if empty.
                # This is much faster than using iterdir.
                with contextlib.suppress(OSError):
                    path_dir.rmdir()
            else:
                problems.add_warning(FileMissingWarning(uid=uid, path=path))
        else:
            problems.add_warning(UIDMissingWarning(uid=uid, path=path))
        if problems:
            return Err(problems)
        return Ok()

    def pending_ids(self) -> list[str]:
        return list(self._pending.keys())

    def pending_paths(self) -> list[Path]:
        return list(self._pending.values())

    def pending_dict(self) -> dict[str, Path]:
        return dict(self._pending)

    @property
    def num_pending(self) -> int:
        return len(self._pending)

    def __contains__(self, uid: str) -> bool:
        return uid in self._pending

    def get_path(self, uid: str) -> Optional[Path]:
        return self._pending.get(uid, None)

    def retrieve(self, uid: str) -> Result[Optional[bytes], Problems]:
        problems = Problems()
        content: Optional[bytes] = None
        path = self._pending.get(uid, None)
        if path:
            if path.exists():
                try:
                    with path.open("rb") as f:
                        content: bytes = f.read()
                except Exception as e:  # pragma: no cover  # noqa: BLE001
                    problems.add_error(e).add_error(
                        ReadFailed("Open or read failed", uid=uid, path=path)
                    )
            else:
                problems.add_error(FileMissing(uid=uid, path=path))
        if problems:
            return Err(problems)
        return Ok(content)

    def reindex(self) -> Result[bool, Problems]:
        problems = Problems()
        self._curr_bytes = 0
        paths: list[_PersistedItem] = []
        last_pat = time.time()
        for base_dir_entry in self._base_dir.iterdir():  # noqa: PLR1702
            try:
                if base_dir_entry.is_dir() and self._is_iso_parseable(base_dir_entry):
                    for day_dir_entry in base_dir_entry.iterdir():
                        if self._pat_watchdog_args:
                            now = time.time()
                            if now > last_pat + self._reindex_pat_seconds:
                                last_pat = now
                                subprocess.run(self._pat_watchdog_args, check=True)  # noqa: S603
                        try:
                            if persisted_item := self._persisted_item_from_file_path(
                                day_dir_entry
                            ):
                                self._curr_bytes += persisted_item.path.stat().st_size
                                paths.append(persisted_item)
                        except Exception as e:  # noqa: BLE001
                            problems.add_error(e).add_error(
                                ReindexError(path=day_dir_entry)
                            )
            except Exception as e:  # noqa: BLE001, PERF203
                problems.add_error(e).add_error(ReindexError())
        # The next line is correct, though PyCharm gives a false-positive warning.
        # paths is a list of tuples, which the dict constructor will treat
        # as a list of key-values pairs, which is the intended behavior.
        self._pending = dict(sorted(paths, key=lambda item: item.path))  # noqa
        if problems:
            return Err(problems)
        return Ok()

    def _today_dir(self) -> Path:
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        return (
            self._base_dir
            / datetime.datetime(
                now.year, now.month, now.day, tzinfo=now.tzinfo
            ).isoformat()
        )

    def _roll_curr_dir(self) -> None:
        today_dir = self._today_dir()
        if today_dir != self._curr_dir:
            self._curr_dir = today_dir
        if not self._curr_dir.exists():
            self._curr_dir.mkdir(parents=True, exist_ok=True)

    @classmethod
    def _make_name(cls, dt: datetime.datetime, uid: str) -> str:
        return f"{dt.isoformat()}.uid[{uid}].json"

    @classmethod
    def _persisted_item_from_file_path(cls, filepath: Path) -> Optional[_PersistedItem]:
        item = None

        try:
            match = cls.FILENAME_RGX.match(filepath.name)
            if match and cls._is_iso_parseable(match.group("dt")):
                item = _PersistedItem(match.group("uid"), filepath)
        except:  # pragma: no cover  # noqa: E722, S110
            pass
        return item

    @classmethod
    def _is_iso_parseable(cls, s: str | Path) -> bool:
        try:
            if isinstance(s, Path):
                s = s.name
            return isinstance(datetime.datetime.fromisoformat(s), datetime.datetime)
        except:  # noqa: E722
            return False



from typing import Any, Self

from pydantic import Field, model_validator
from pydantic_settings import SettingsConfigDict

from gwproactor import ProactorSettings
from gwproactor.config import Paths
from gwproactor_test.dummies.names import DUMMY_ADMIN_NAME, DUMMY_ADMIN_SHORT_NAME
from gwproactor_test.dummies.tree.link_settings import TreeLinkSettings


class AdminLinkSettings(TreeLinkSettings):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            client_name=DUMMY_ADMIN_NAME,
            long_name=DUMMY_ADMIN_NAME,
            short_name=DUMMY_ADMIN_SHORT_NAME,
            **kwargs,
        )


class DummyAdminSettings(ProactorSettings):
    target_gnode: str = ""
    paths: Paths = Field({}, validate_default=True)
    link: AdminLinkSettings = AdminLinkSettings()
    model_config = SettingsConfigDict(env_prefix="GWADMIN_", env_nested_delimiter="__")

    @model_validator(mode="before")
    @classmethod
    def pre_root_validator(cls, values: dict) -> dict:
        return ProactorSettings.update_paths_name(values, DUMMY_ADMIN_NAME)

    @model_validator(mode="after")
    def validate(self) -> Self:
        self.link.update_tls_paths(self.paths.certs_dir, self.link.client_name)
        return self



"""Settings for the GridWorks Scada, readable from environment and/or from env files."""

from gwproactor.config.logging import (
    DEFAULT_BYTES_PER_LOG_FILE,
    DEFAULT_FRACTIONAL_SECOND_FORMAT,
    DEFAULT_LOG_FILE_NAME,
    DEFAULT_LOGGING_FORMAT,
    DEFAULT_NUM_LOG_FILES,
    FormatterSettings,
    LoggerLevels,
    LoggingSettings,
    RotatingFileHandlerSettings,
)
from gwproactor.config.mqtt import MQTTClient
from gwproactor.config.paths import (
    DEFAULT_BASE_DIR,
    DEFAULT_BASE_NAME,
    DEFAULT_LAYOUT_FILE,
    DEFAULT_NAME,
    DEFAULT_NAME_DIR,
    Paths,
)
from gwproactor.config.proactor_settings import ProactorSettings

DEFAULT_ENV_FILE = ".env"

__all__ = [
    "DEFAULT_BASE_DIR",
    "DEFAULT_BASE_NAME",
    "DEFAULT_BYTES_PER_LOG_FILE",
    # paths
    "DEFAULT_ENV_FILE",
    "DEFAULT_FRACTIONAL_SECOND_FORMAT",
    "DEFAULT_LAYOUT_FILE",
    # logging
    "DEFAULT_LOGGING_FORMAT",
    "DEFAULT_LOG_FILE_NAME",
    "DEFAULT_NAME",
    "DEFAULT_NAME_DIR",
    "DEFAULT_NUM_LOG_FILES",
    "FormatterSettings",
    "LoggerLevels",
    "LoggingSettings",
    # mqtt
    "MQTTClient",
    "Paths",
    # proactor
    "ProactorSettings",
    "RotatingFileHandlerSettings",
]



import typing
from typing import Optional

from gwproto import Message, MQTTCodec, MQTTTopic, create_message_model

from gwproactor import ProactorSettings
from gwproactor.links import QOS
from gwproactor.links.link_settings import LinkSettings
from gwproactor.persister import TimedRollingFilePersister
from gwproactor.proactor_implementation import Proactor
from gwproactor_test.dummies.names import (
    CHILD_SHORT_NAME,
    DUMMY_CHILD_NAME,
    DUMMY_PARENT_NAME,
    PARENT_SHORT_NAME,
)
from gwproactor_test.dummies.pair.child_config import DummyChildSettings


class ChildMQTTCodec(MQTTCodec):
    def __init__(self) -> None:
        super().__init__(
            create_message_model(
                "ChildMessageDecoder",
                [
                    "gwproto.messages",
                    "gwproactor.message",
                ],
            )
        )

    def validate_source_and_destination(self, src: str, dst: str) -> None:
        if src != DUMMY_PARENT_NAME or dst != CHILD_SHORT_NAME:
            raise ValueError(
                "ERROR validating src and/or dst\n"
                f"  exp: {DUMMY_PARENT_NAME} -> {CHILD_SHORT_NAME}\n"
                f"  got: {src} -> {dst}"
            )


class DummyChild(Proactor):
    PARENT_MQTT = "gridworks"

    def __init__(
        self,
        name: str = "",
        settings: Optional[DummyChildSettings] = None,
    ) -> None:
        super().__init__(
            name=name or DUMMY_CHILD_NAME,
            settings=DummyChildSettings() if settings is None else settings,
        )
        self._links.add_mqtt_link(
            LinkSettings(
                client_name=DummyChild.PARENT_MQTT,
                gnode_name=DUMMY_PARENT_NAME,
                spaceheat_name=PARENT_SHORT_NAME,
                mqtt=settings.parent_mqtt,
                codec=ChildMQTTCodec(),
                upstream=True,
            )
        )
        for topic in [
            MQTTTopic.encode_subscription(Message.type_name(), "1", "a"),
            MQTTTopic.encode_subscription(Message.type_name(), "2", "b"),
        ]:
            self._links.subscribe(self.PARENT_MQTT, topic, QOS.AtMostOnce)
        self._links.log_subscriptions("construction")

    @classmethod
    def make_event_persister(
        cls, settings: ProactorSettings
    ) -> TimedRollingFilePersister:
        return TimedRollingFilePersister(settings.paths.event_dir)

    @property
    def publication_name(self) -> str:
        return self.name

    @property
    def subscription_name(self) -> str:
        return CHILD_SHORT_NAME

    @property
    def settings(self) -> DummyChildSettings:
        return typing.cast(DummyChildSettings, self._settings)



import functools
from dataclasses import dataclass
from typing import Any, Callable, Optional

from gwproactor.links.timer_interface import TimerManagerInterface


@dataclass
class AckWaitInfo:
    link_name: str
    message_id: str
    timer_handle: Any
    context: Any = None


AckTimerCallback = Callable[[AckWaitInfo], None]

DEFAULT_ACK_DELAY = 5.0


class AckManager:
    _acks: dict[str, dict[str, AckWaitInfo]]
    _timer_mgr: TimerManagerInterface
    _callback: AckTimerCallback
    _default_delay_seconds: float

    def __init__(
        self,
        timer_mgr: TimerManagerInterface,
        callback: AckTimerCallback,
        delay: Optional[float] = DEFAULT_ACK_DELAY,
    ) -> None:
        self._acks = {}
        self._timer_mgr = timer_mgr
        self._user_callback = callback
        self._default_delay_seconds = delay

    def start_ack_timer(
        self,
        link_name: str,
        message_id: str,
        context: Optional[Any] = None,
        delay_seconds: Optional[float] = None,
    ) -> AckWaitInfo:
        self.cancel_ack_timer(link_name, message_id)
        delay_seconds = (
            self._default_delay_seconds if delay_seconds is None else delay_seconds
        )
        wait_info = AckWaitInfo(
            link_name=link_name,
            message_id=message_id,
            timer_handle=self._timer_mgr.start_timer(
                self._default_delay_seconds if delay_seconds is None else delay_seconds,
                functools.partial(self._timeout, link_name, message_id),
            ),
            context=context,
        )
        if link_name not in self._acks:
            self._acks[link_name] = {}
        self._acks[link_name][message_id] = wait_info
        return wait_info

    def add_link(self, link_name: str) -> None:
        self._acks[link_name] = {}

    def _pop_wait_info(self, link_name: str, message_id: str) -> Optional[AckWaitInfo]:
        if (client_acks := self._acks.get(link_name, None)) is not None:
            return client_acks.pop(message_id, None)
        return None

    def _timeout(self, link_name: str, message_id: str) -> None:
        if (wait_info := self._pop_wait_info(link_name, message_id)) is not None:
            self._user_callback(wait_info)

    def cancel_ack_timer(self, link_name: str, message_id: str) -> AckWaitInfo:
        if (wait_info := self._pop_wait_info(link_name, message_id)) is not None:
            self._timer_mgr.cancel_timer(wait_info.timer_handle)
        return wait_info

    def cancel_ack_timers(self, link_name: str) -> list[AckWaitInfo]:
        if link_name in self._acks:
            wait_infos = self._acks[link_name]
            self._acks[link_name] = {}
            for wait_info in wait_infos.values():
                self._timer_mgr.cancel_timer(wait_info.timer_handle)
        else:
            wait_infos = []
        return wait_infos

    def num_acks(self, link_name: str) -> int:
        if (client_acks := self._acks.get(link_name, None)) is not None:
            return len(client_acks)
        return 0

    @property
    def default_delay_seconds(self) -> float:
        return self._default_delay_seconds



# ruff: noqa: ERA001
import dataclasses
from abc import abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Callable, Optional, Tuple, Type, TypeVar, cast

from gwproto import Message
from gwproto.messages import CommEvent, EventBase, EventT, PingMessage
from paho.mqtt.client import MQTT_ERR_SUCCESS, MQTTMessageInfo

from gwproactor import Proactor, ProactorSettings, Runnable, ServicesInterface
from gwproactor.config import LoggerLevels
from gwproactor.links import LinkManager, MQTTClients, MQTTClientWrapper
from gwproactor.message import (
    DBGCommands,
    DBGPayload,
    MQTTReceiptPayload,
    MQTTSubackPayload,
)
from gwproactor.stats import LinkStats, ProactorStats


def split_subscriptions(client_wrapper: MQTTClientWrapper) -> Tuple[int, Optional[int]]:
    for topic, qos in client_wrapper.subscription_items():
        MQTTClientWrapper.subscribe(client_wrapper, topic, qos)
    return MQTT_ERR_SUCCESS, None


@dataclass
class RecorderLinkStats(LinkStats):
    comm_events: list[CommEvent] = field(default_factory=list)
    forwarded: dict[str, int] = field(default_factory=lambda: defaultdict(int))
    event_counts: dict[str, dict[str, int]] = field(
        default_factory=lambda: defaultdict(lambda: defaultdict(int))
    )

    def __str__(self) -> str:
        s = super().__str__()
        if self.comm_events:
            s += "\n  Comm events:"
            for comm_event in self.comm_events:
                copy_event = comm_event.model_copy(
                    update={"MessageId": comm_event.MessageId[:6] + "..."}
                )
                s += f"\n    {str(copy_event)[:154]}"
        if self.forwarded:
            s += "\n  Forwarded events *sent* by type:"
            for message_type in sorted(self.forwarded):
                s += f"\n    {self.forwarded[message_type]:3d}: [{message_type}]"
        if self.event_counts:
            s += "\n  Events *received* by src and type:"
            for event_src in sorted(self.event_counts):
                s += f"\n    src: {event_src}"
                forwards_from_src = self.event_counts[event_src]
                for message_type in sorted(forwards_from_src):
                    s += f"\n      {forwards_from_src[message_type]:3d}: [{message_type}]"
        return s


class RecorderStats(ProactorStats):
    @classmethod
    def make_link(cls, link_name: str) -> RecorderLinkStats:
        return RecorderLinkStats(link_name)


ProactorT = TypeVar("ProactorT", bound=Proactor)


class RecorderInterface(ServicesInterface, Runnable):
    @classmethod
    @abstractmethod
    def make_stats(cls) -> RecorderStats: ...

    @abstractmethod
    def split_client_subacks(self, client_name: str) -> None: ...

    @abstractmethod
    def restore_client_subacks(self, client_name: str) -> None: ...

    @abstractmethod
    def pause_subacks(self) -> None: ...

    @abstractmethod
    def release_subacks(self, num_released: int = -1) -> None: ...

    @abstractmethod
    def force_ping(self, client_name: str) -> None: ...

    @abstractmethod
    def summary_str(self) -> None: ...

    @abstractmethod
    def summarize(self) -> None: ...

    @property
    @abstractmethod
    def mqtt_clients(self) -> MQTTClients: ...

    @abstractmethod
    def mqtt_client_wrapper(self, client_name: str) -> MQTTClientWrapper: ...

    @abstractmethod
    def mqtt_subscriptions(self, client_name: str) -> list[str]: ...

    @abstractmethod
    def disable_derived_events(self) -> None: ...

    @abstractmethod
    def enable_derived_events(self) -> None: ...

    @abstractmethod
    def mqtt_quiescent(self) -> bool: ...


@dataclass
class _PausedAck:
    link_name: str
    message: Message
    qos: int
    context: Optional[Any]


class RecorderLinks(LinkManager):
    acks_paused: bool
    needs_ack: list[_PausedAck]

    # noinspection PyMissingConstructor
    def __init__(self, other: LinkManager) -> None:
        self.__dict__ = other.__dict__
        self.acks_paused = False
        self.needs_ack = []

    def publish_message(
        self, client: str, message: Message, qos: int = 0, context: Any = None
    ) -> MQTTMessageInfo:
        if self.acks_paused:
            self.needs_ack.append(_PausedAck(client, message, qos, context))
            return MQTTMessageInfo(-1)
        # noinspection PyProtectedMember
        return super().publish_message(client, message, qos=qos, context=context)

    def release_acks(self, clear: bool = False, num_to_release: int = -1) -> int:
        # self._logger.info(
        #     f"++release_acks: clear:{clear}  num_to_release:{num_to_release}"
        # )
        # path_dbg = 0
        if clear or num_to_release < 1:
            # path_dbg |= 0x00000001
            self.acks_paused = False
            needs_ack = self.needs_ack
            self.needs_ack = []
        else:
            # path_dbg |= 0x00000002
            num_to_release = min(num_to_release, len(self.needs_ack))
            needs_ack = self.needs_ack[:num_to_release]
            self.needs_ack = self.needs_ack[num_to_release:]
            # self._logger.info(f"needs_ack: {needs_ack}")
            # self._logger.info(f"self.needs_ack: {self.needs_ack}")
        if not clear:
            # path_dbg |= 0x00000004
            for paused_ack in needs_ack:
                # path_dbg |= 0x00000008
                super().publish_message(**dataclasses.asdict(paused_ack))  # noqa
        # self._logger.info(
        #     f"--release_acks: clear:{clear}  num_to_release:{num_to_release}  path:0x{path_dbg:08X}"
        # )
        return len(needs_ack)

    def generate_event(self, event: EventT) -> None:
        if not event.Src:
            event.Src = self.publication_name
        if isinstance(event, CommEvent) and event.Src == self.publication_name:
            cast(
                RecorderLinkStats, self._stats.link(event.PeerName)
            ).comm_events.append(event)
        if event.Src != self.publication_name:
            cast(RecorderLinkStats, self._stats.link(event.Src)).forwarded[
                event.TypeName
            ] += 1
        super().generate_event(event)


def make_recorder_class(  # noqa: C901
    proactor_type: Type[ProactorT],
) -> Callable[..., RecorderInterface]:
    class Recorder(proactor_type):
        _subacks_paused: dict[str, bool]
        _subacks_available: dict[str, list[Message]]
        _mqtt_messages_dropped: dict[str, bool]

        def __init__(
            self, name: str, settings: ProactorSettings, **kwargs_: Any
        ) -> None:
            super().__init__(name=name, settings=settings, **kwargs_)
            self._subacks_paused = defaultdict(bool)
            self._subacks_available = defaultdict(list)
            self._mqtt_messages_dropped = defaultdict(bool)
            self._links = RecorderLinks(self._links)

        @classmethod
        def make_stats(cls) -> RecorderStats:
            return RecorderStats()

        @property
        def needs_ack(self) -> list[_PausedAck]:
            return self._links.needs_ack

        def subacks_paused(self, client_name: str) -> bool:
            return self._subacks_paused[client_name]

        def num_subacks_available(self, client_name: str) -> int:
            return len(self._subacks_available[client_name])

        def clear_subacks(self, client_name: str) -> None:
            self._subacks_available[client_name] = []

        def mqtt_messages_dropped(self, client_name: str) -> bool:
            return self._mqtt_messages_dropped[client_name]

        def upstream_subacks_paused(self) -> bool:
            return self.subacks_paused(self.upstream_client)

        def num_upstream_subacks_available(self) -> int:
            return self.num_subacks_available(self.upstream_client)

        def clear_upstream_subacks(self) -> None:
            self._subacks_available[self.upstream_client] = []

        def upstream_mqtt_messages_dropped(self) -> bool:
            return self.mqtt_messages_dropped(self.upstream_cleint)

        def split_client_subacks(self: ProactorT, client_name: str) -> None:
            client_wrapper = self.mqtt_client_wrapper(client_name)

            def member_split_subscriptions() -> Tuple[int, Optional[int]]:
                return split_subscriptions(client_wrapper)

            client_wrapper.subscribe_all = member_split_subscriptions

        def restore_client_subacks(self: ProactorT, client_name: str) -> None:
            client_wrapper = self.mqtt_client_wrapper(client_name)
            client_wrapper.subscribe_all = MQTTClientWrapper.subscribe_all

        def pause_subacks(self, client_name: str) -> None:
            self._subacks_paused[client_name] = True

        def pause_upstream_subacks(self) -> None:
            self.pause_subacks(self.upstream_client)

        def release_subacks(
            self: ProactorT, client_name: str, num_released: int = -1
        ) -> None:
            self._subacks_paused[client_name] = False
            if num_released < 0:
                num_released = len(self._subacks_available[client_name])
            release = self._subacks_available[client_name][:num_released]
            remaining = self._subacks_available[client_name][num_released:]
            self._subacks_available[client_name] = remaining
            for message in release:
                self._receive_queue.put_nowait(message)

        def release_upstream_subacks(self: ProactorT, num_released: int = -1) -> None:
            self.release_subacks(self.upstream_client, num_released)

        async def process_message(self, message: Message) -> None:
            if (
                isinstance(message.Payload, MQTTSubackPayload)
                and self._subacks_paused[message.Payload.client_name]
            ):
                self._subacks_available[message.Payload.client_name].append(message)
            else:
                await super().process_message(message)

        def _derived_process_mqtt_message(
            self, message: Message[MQTTReceiptPayload], decoded: Message[Any]
        ) -> None:
            super()._derived_process_mqtt_message(message, decoded)  # noqa
            match decoded.Payload:
                case EventBase() as event:
                    stats = cast(
                        RecorderLinkStats, self._stats.link(message.Payload.client_name)
                    )
                    stats.event_counts[event.Src][event.TypeName] += 1

        def pause_acks(self) -> None:
            self._links.acks_paused = True

        def release_acks(self, clear: bool = False, num_to_release: int = -1) -> int:
            return self._links.release_acks(clear, num_to_release=num_to_release)

        def set_ack_timeout_seconds(self, delay: float) -> None:
            self.links.ack_manager._default_delay_seconds = delay  # noqa: SLF001

        def drop_mqtt(self, client_name: str, drop: bool) -> None:
            self._mqtt_messages_dropped[client_name] = drop

        def _process_mqtt_message(self, message: Message[MQTTReceiptPayload]) -> None:
            if not self._mqtt_messages_dropped[message.Payload.client_name]:
                # noinspection PyProtectedMember
                super()._process_mqtt_message(message)

        def summary_str(self: ProactorT) -> str:
            s = str(self.stats)
            s += "\nLink states:\n"
            for link_name in self.stats.links:
                s += f"  {link_name:10s}  {self._links.link_state(link_name).value}\n"
            s += self.links.subscription_str().lstrip()
            s += "Pending acks:\n"
            for link_name in self.stats.links:
                s += f"  {link_name:10s}  {self._links.num_acks(link_name):3d}\n"
            s += self._links.get_reuploads_str() + "\n"
            s += "Paused Subacks:"
            for link_name in self.stats.links:
                s += (
                    f"  {link_name:10s}  "
                    f"subacks paused: {self._subacks_paused[link_name]}  "
                    f"subacks available: {len(self._subacks_available[link_name])}\n"
                )
            return s

        def summarize(self: ProactorT) -> None:
            self._logger.info(self.summary_str())

        def force_ping(self, client_name: str) -> None:
            self._links.publish_message(
                client_name, PingMessage(Src=self.publication_name)
            )

        @property
        def mqtt_clients(self) -> MQTTClients:
            return self._links.mqtt_clients()

        def mqtt_client_wrapper(self, client_name: str) -> MQTTClientWrapper:
            return self._links.mqtt_client_wrapper(client_name)

        def mqtt_subscriptions(self, client_name: str) -> list[str]:
            return [
                item[0]
                for item in self.mqtt_client_wrapper(client_name).subscription_items()
            ]

        def all_mqtt_subscriptions(self) -> list[str]:
            subscriptions = []
            for client_name in self.mqtt_clients.clients:
                subscriptions.extend(self.mqtt_subscriptions(client_name))
            return subscriptions

        def send_dbg(
            self,
            client_name: str,
            message_summary: int = -1,
            lifecycle: int = -1,
            comm_event: int = -1,
            command: Optional[DBGCommands | str] = None,
        ) -> None:
            if isinstance(command, str):
                command = DBGCommands(command)
            self.send_threadsafe(
                Message(
                    Src=self.name,
                    Dst=client_name,
                    Payload=DBGPayload(
                        Levels=LoggerLevels(
                            message_summary=message_summary,
                            lifecycle=lifecycle,
                            comm_event=comm_event,
                        ),
                        Command=command,
                    ),
                )
            )

        def _derived_process_message(self, message: Message) -> None:
            match message.Payload:
                case DBGPayload():
                    message.Header.Src = self.publication_name
                    dst_client = message.Header.Dst
                    message.Header.Dst = ""
                    self._links.publish_message(dst_client, message)
                case _:
                    # noinspection PyProtectedMember
                    super()._derived_process_message(message)

        def mqtt_quiescent(self) -> bool:
            if hasattr(super(), "mqtt_quiescent"):
                return super().mqtt_quiescent()
            return self._links.link(self.upstream_client).active_for_send()

        def _call_super_if_present(self, function_name: str) -> None:
            if hasattr(super(), function_name):
                getattr(super(), function_name)()

        def disable_derived_events(self) -> None:
            self._call_super_if_present("disable_dervived_events")

        def enable_derived_events(self) -> None:
            self._call_super_if_present("enable_dervived_events")

    return Recorder
